#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.db import standard_attr


# TODO(boden): remove shims when all consumers use lib for this module
StandardAttribute = standard_attr.StandardAttribute
HasStandardAttributes = standard_attr.HasStandardAttributes
get_standard_attr_resource_model_map = (
    standard_attr.get_standard_attr_resource_model_map)
get_tag_resource_parent_map = standard_attr.get_tag_resource_parent_map
